#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2008 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of kowitest for details.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import os
import re

from Ft.Xml      import Parse, InputSource, EMPTY_NAMESPACE
from Ft.Lib.Uri  import OsPathToUri 
from Ft.Xml.Xslt import Processor

from Ft.Xml.cDomlette import Text # in newer versions found in Domlette

from server import MyHTTPServer, Response

NO_SECURE_ID       = re.compile("[^_a-zA-Z0-9\-]")
VALID_TEST_RESULTS = set(["passed", "failed", ""])

DEFAULT_PORT = 8080

XFORMS = {}

def transform_file(xform):
    """returns absolute pathname of transformation file."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "xforms", xform))

def get_first(params, key):
    """helper to extract first entry of value list in a dictionary"""
    p = params.get(key)
    if p: return p[0]
    return None

def node_text(node):
    """extract all text children for a given node and concat their
       content a a single string"""
    parts = []
    for child in node.childNodes:
        if isinstance(child, Text):
            parts.append(child.nodeValue)
    return ''.join(parts)

def get_transform(xform):
    """fetches a XSL transformation and caches it to reduce I/O and parsing"""
    try:
        return XFORMS[xform]
    except KeyError:
        fname     = transform_file(xform)
        uri       = OsPathToUri(fname)
        transform = Parse(InputSource.DefaultFactory.fromUri(uri))
        XFORMS[xform] = transform
        return transform

class Server:

    def __init__(self, filename, port = DEFAULT_PORT):
        self.filename = filename
        self.port     = port

    def run(self):
        """setup the mapping of the web server and run it"""
        self.doc = Parse(self.filename)
        mappings = {
            'index'           : self.index,
            'index.html'      : self.index,
            ''                : self.index,
            'saveheader'      : self.saveheader,
            'quit'            : self.quit,
            'select-testsuite': self.select_testsuite,
            'set-test'        : self.set_test,
            'execute'         : self.execute_script
        }

        serverAddress = ('', self.port)
        self.httpd = MyHTTPServer(serverAddress, mappings)

        self.done = False
        while not self.done:
            self.httpd.handle_request()

    def get_node(self, parent, name):
        """get a named child node from parent node
           or creates it if does not exist."""
        child = parent.xpath(name)
        if not child:
            child = self.doc.createElementNS(EMPTY_NAMESPACE, name)
            parent.appendChild(child)
        else:
            child = child[0]
        return child

    def replace_text(self, node, txt):
        """replaces the first text child with a new one."""
        text = self.doc.createTextNode(txt)
        for child in node.childNodes:
            if isinstance(child, Text):
                node.replaceChild(text, child)
                return
        node.appendChild(text)

    def index(self, path, params):
        """render the overview"""
        processor = Processor.Processor()
        xform = get_transform("overview.xsl")
        processor.appendStylesheetNode(xform)
        return Response(processor.runNode(self.doc))

    def quit(self, path, params):
        """shutdown the server"""
        self.done = True
        return Response("<html><body>Server is gone!</body></html>")
        # self.httpd.shutdown() # python 2.6
        # sys.exit(0)           # too hard

    def get_meta(self):
        """helper to get the meta node of the document"""
        meta = self.doc.xpath("/testplan/meta")
        return meta and meta[0] or None

    def store_document(self):
        """stores the document back to disk. 
           TODO: We really need a Null transform to presere CDATA chunks
           which are importing for the embedded scripts."""
        f = None
        try:
            xform = get_transform("identity.xsl")
            f = open(self.filename, "w")
            processor = Processor.Processor()
            processor.appendStylesheetNode(xform)
            processor.runNode(self.doc, outputStream = f)
        finally:
            if f:
                try: f.close()
                except: pass

    def render_testsuite(self, name):
        """render a specified test suite"""
        processor = Processor.Processor()
        xform = get_transform("testsuite.xsl")
        processor.appendStylesheetNode(xform)
        return Response(
            processor.runNode(
                self.doc,
                topLevelParams = {u'selected-suite': name}))

    def select_testsuite(self, path, params):
        """selects a test suite from overview and renders it."""
        suite = get_first(params, 'suite')
        if not suite: return self.index(path, params)
        suite = NO_SECURE_ID.sub('', suite)
        if not self.doc.xpath("//testsuite[@id='%s']" % suite):
            return self.index(path, params)
        return self.render_testsuite(suite)

    def execute_script(self, path, params):
        """executes a script node"""
        suite  = get_first(params, 'suite')
        test   = get_first(params, 'test')
        script = get_first(params, 'script')

        if suite  is None \
        or test   is None \
        or script is None \
        or not test.isdigit() \
        or not script.isdigit():
            return self.index(path, params)

        suite = NO_SECURE_ID.sub('', suite)

        res = self.doc.xpath(
            "//testsuite[@id='%s']"
            "/test[count(preceding-sibling::*[name()='test'])=%s]"
            "/steps/script[count(preceding-sibling::*[name()='script'])=%s]" % (
            suite, test, script))

        if not res: return self.index(path, params)

        res = res[0]

        if res.getAttributeNS(EMPTY_NAMESPACE, 'language') == 'python':
            script_text = node_text(res)
            exec(script_text) # !!!XXX!!! Security risk

        return self.render_testsuite(suite)

    def set_test(self, path, params):
        """toggles a given result to a given value"""
        suite  = get_first(params, 'suite')
        test   = get_first(params, 'test')
        result = get_first(params, 'result')
        value  = get_first(params, 'value')

        if suite  is None \
        or test   is None \
        or result is None \
        or value  is None \
        or not test.isdigit() \
        or not result.isdigit() \
        or value not in VALID_TEST_RESULTS: 
            return self.index(path, params)

        suite = NO_SECURE_ID.sub('', suite)

        res = self.doc.xpath(
            "//testsuite[@id='%s']"
            "/test[count(preceding-sibling::*[name()='test'])=%s]"
            "/results/result[count(preceding-sibling::*[name()='result'])=%s]" % (
            suite, test, result))

        if not res: return self.index(path, params)

        res[0].setAttributeNS(EMPTY_NAMESPACE, 'value', value)

        self.store_document()

        return self.render_testsuite(suite)
        
    def saveheader(self, path, params):
        """modifies the meta data of a test document"""
        meta = self.get_meta()
        if meta:
            for name in (u"date", u"test-system", u"tester", u"test-duration", u"comment"):
                param = get_first(params, name)
                if not param is None:
                    node = self.get_node(meta, name)
                    self.replace_text(node, param)

        self.store_document()
        return self.index(path, params)

def main():
    if len(sys.argv) < 2:
        print >> sys.stderr, "Need XML test file."
        sys.exit(1)

    port = DEFAULT_PORT

    if len(sys.argv) > 2: port = int(sys.argv[2])

    server = Server(sys.argv[1], port)
    print "HTTP server listening at http://localhost:%d/" % port
    server.run()

if __name__ == '__main__':
    main()
